﻿using System.Collections.Generic;
using System.Linq;
using BenchmarkDotNet.Attributes;
using HeuristicLab.Common;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Problems.DataAnalysis.Symbolic;
using HeuristicLab.Random;

namespace HeuristicLab.Benchmarks {
  [ClrJob]
  [MinColumn, MaxColumn, MeanColumn, MedianColumn]
  public class InterpreterPerformance {
    private readonly ISymbolicDataAnalysisExpressionTreeInterpreter nativeInterpreter = new SymbolicDataAnalysisExpressionTreeNativeInterpreter();
    private readonly ISymbolicDataAnalysisExpressionTreeInterpreter batchInterpreter = new SymbolicDataAnalysisExpressionTreeBatchInterpreter();

    private readonly Dataset dataset;
    private readonly ISymbolicDataAnalysisGrammar grammar = new ArithmeticExpressionGrammar();
    private ISymbolicExpressionTree[] trees;
    //private int[] rows;
    private IEnumerable<int> rows;

    private const int Columns = 50;
    private const int popSize = 1000;
    private const int minSize = 1;
    private const int maxSize = 100;
    private readonly int totalNodes;

    private readonly MersenneTwister mt = new MersenneTwister(1234);

    private static IEnumerable<int> IntRange(int start, int end, int step, bool closed = true) {
      return SequenceGenerator.GenerateSteps((decimal)start, end, step, closed).Select(x => (int)x);
    }

    public InterpreterPerformance() {
      int maxRows = RowsValues.Max();
      dataset = Util.CreateRandomDataset(mt, maxRows, Columns);
      trees = Util.CreateRandomTrees(mt, dataset, grammar, popSize, minSize, maxSize, 0, 0);

      totalNodes = trees.Sum(x => x.Length);
    }

    [ParamsSource(nameof(TotalNodesValues))]
    public int TotalNodes { get; set; }

    [ParamsSource(nameof(RowsValues))]
    public int Rows { get; set; }

    //public IEnumerable<int> RowsValues => IntRange(1_000, 9_000, 1_000).Concat(IntRange(10_000, 50_000, 10_000));
    public IEnumerable<int> RowsValues => new[] { 5000 };
    public IEnumerable<int> TotalNodesValues => new[] { totalNodes };

    [GlobalSetup]
    public void Setup() {
      rows = Enumerable.Range(0, Rows);
    }

    [Benchmark]
    public int NativeInterpreter() => TestInterpreter(nativeInterpreter);

    [Benchmark]
    public int BatchInterpreter() => TestInterpreter(batchInterpreter);

    #region helper methods
    private int TestInterpreter(ISymbolicDataAnalysisExpressionTreeInterpreter interpreter) {
      int count = 0;
      foreach (var tree in trees) {
        count += interpreter.GetSymbolicExpressionTreeValues(tree, dataset, rows).Count();
      }
      return count;
    }
    #endregion
  }
}
